@@ -54,6 +54,7 @@ OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. */
 #include "player.h"
 #include "playerops.h"
 #include "playlist.h"
+#include "search_ui.h"
 #include "settings.h"
 #include "sound.h"
 #include "soundcommon.h"
@@ -61,7 +62,6 @@ OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. */
 #include "utils.h"
 
 // #define DEBUG 1
-#define MAX_SEQ_LEN 1024    // Maximum length of sequence buffer
 #define MAX_TMP_SEQ_LEN 256 // Maximum length of temporary sequence buffer
 #define COOLDOWN_MS 500
 #define COOLDOWN2_MS 100
@@ -72,12 +72,14 @@ static bool eventProcessed = false;
 char digitsPressed[MAX_SEQ_LEN];
 int digitsPressedCount = 0;
 int maxDigitsPressedCount = 9;
+static unsigned int updateCounter = 0;
 bool gPressed = false;
 bool loadingAudioData = false;
 bool goingToSong = false;
 bool startFromTop = false;
 bool exactSearch = false;
 AppSettings settings;
+int fuzzySearchThreshold = 2;
 
 bool isCooldownElapsed(int milliSeconds)
 {
@@ -93,7 +95,7 @@ struct Event processInput()
 {
         struct Event event;
         event.type = EVENT_NONE;
-        event.key = '\0';
+        event.key[0] = '\0';
         bool cooldownElapsed = false;
         bool cooldown2Elapsed = false;
 
@@ -122,7 +124,7 @@ struct Event processInput()
                 seqLength = seqLength + readInputSequence(tmpSeq, sizeof(tmpSeq));
 
                 // Release most keys directly, seekbackward and seekforward can be read continuously
-                if (seqLength <= 0 && strcmp(seq, settings.seekBackward) != 0 && strcmp(seq, settings.seekForward) != 0)
+                if (seqLength <= 0 && strcmp(seq + 1, settings.seekBackward) != 0 && strcmp(seq + 1, settings.seekForward) != 0)
                 {
                         keyReleased = 1;
                         break;
@@ -135,16 +137,11 @@ struct Event processInput()
 
                 strcat(seq, tmpSeq);
 
-                c_sleep(10);
-
                 // This slows the continous reads down to not get a a too fast scrolling speed
-                if (strcmp(seq, settings.hardScrollUp) == 0 || strcmp(seq, settings.hardScrollDown) == 0 || strcmp(seq, settings.scrollUpAlt) == 0 ||
-                    strcmp(seq, settings.scrollDownAlt) == 0 || strcmp(seq, settings.seekBackward) == 0 || strcmp(seq, settings.seekForward) == 0 ||
-                    strcmp(seq, settings.hardNextPage) == 0 || strcmp(seq, settings.hardPrevPage) == 0)
+                if (strcmp(seq + 1, settings.hardScrollUp) == 0 || strcmp(seq + 1, settings.hardScrollDown) == 0 || strcmp(seq + 1, settings.scrollUpAlt) == 0 ||
+                    strcmp(seq + 1, settings.scrollDownAlt) == 0 || strcmp(seq + 1, settings.seekBackward) == 0 || strcmp(seq + 1, settings.seekForward) == 0 ||
+                    strcmp(seq + 1, settings.hardNextPage) == 0 || strcmp(seq + 1, settings.hardPrevPage) == 0)
                 {
-                        // Do dummy reads to prevent scrolling continuing after we release the key
-                        readInputSequence(tmpSeq, 3);
-                        readInputSequence(tmpSeq, 3);
                         keyReleased = 0;
                         break;
                 }
@@ -157,7 +154,26 @@ struct Event processInput()
 
         eventProcessed = true;
         event.type = EVENT_NONE;
-        event.key = seq[0];
+
+        strncpy(event.key, seq, MAX_SEQ_LEN);
+
+        if (appState.currentView == SEARCH_VIEW)
+        {
+                if (strcmp(event.key, "\x7F") == 0 || strcmp(event.key, "\x08") == 0)
+                {
+                        removeFromSearchText(getLibrary());
+                        chosenSearchResultRow = 0;
+                        fuzzySearch(getLibrary(), fuzzySearchThreshold);
+                        event.type = EVENT_SEARCH;
+                }
+                else if (((strlen(event.key) == 1 && event.key[0] != '\033' && event.key[0] != '\n' && event.key[0] != '\r') || strcmp(event.key, " ") == 0 || (unsigned char)event.key[0] >= 0xC0))
+                {
+                        addToSearchText(event.key);
+                        chosenSearchResultRow = 0;
+                        fuzzySearch(getLibrary(), fuzzySearchThreshold);
+                        event.type = EVENT_SEARCH;
+                }
+        }         
 
         // Map keys to events
         EventMapping keyMappings[] = {{settings.scrollUpAlt, EVENT_SCROLLPREV},
@@ -169,6 +185,7 @@ struct Event processInput()
                                       {settings.volumeDown, EVENT_VOLUME_DOWN},
                                       {settings.togglePause, EVENT_PLAY_PAUSE},
                                       {settings.quit, EVENT_QUIT},
+                                      {settings.hardQuit, EVENT_QUIT},
                                       {settings.toggleShuffle, EVENT_SHUFFLE},
                                       {settings.toggleCovers, EVENT_TOGGLECOVERS},
                                       {settings.toggleVisualizer, EVENT_TOGGLEVISUALIZER},
@@ -189,6 +206,7 @@ struct Event processInput()
                                       {settings.hardScrollDown, EVENT_SCROLLNEXT},
                                       {settings.hardShowInfo, EVENT_SHOWINFO},
                                       {settings.hardShowInfoAlt, EVENT_SHOWINFO},
+                                      {settings.hardShowSearch, EVENT_SHOWSEARCH},
                                       {settings.hardShowKeys, EVENT_SHOWKEYBINDINGS},
                                       {settings.hardShowKeysAlt, EVENT_SHOWKEYBINDINGS},
                                       {settings.hardEndOfPlaylist, EVENT_GOTOENDOFPLAYLIST},
@@ -196,24 +214,32 @@ struct Event processInput()
                                       {settings.hardShowTrackAlt, EVENT_SHOWTRACK},
                                       {settings.hardShowLibrary, EVENT_SHOWLIBRARY},
                                       {settings.hardShowLibraryAlt, EVENT_SHOWLIBRARY},
+                                      {settings.hardShowSearchAlt,  EVENT_SHOWSEARCH},
                                       {settings.hardNextPage, EVENT_NEXTPAGE},
                                       {settings.hardPrevPage, EVENT_PREVPAGE},
                                       {settings.hardRemove, EVENT_REMOVE}};
 
-        int numKeyMappings = sizeof(keyMappings) / sizeof(EventMapping);
+        int numKeyMappings = sizeof(keyMappings) / sizeof(EventMapping);             
 
         // Set event for pressed key
         for (int i = 0; i < numKeyMappings; i++)
         {
-                if (strcmp(seq, keyMappings[i].seq) == 0)
+                if (keyMappings[i].seq[0] != '\0' &&
+                    ((seq[0] == '\033' && strlen(seq) > 1 && strcmp(seq + 1, keyMappings[i].seq) == 0) ||
+                     strcmp(seq, keyMappings[i].seq) == 0))
                 {
+                        if (event.type == EVENT_SEARCH && keyMappings[i].eventType != EVENT_GOTOSONG)
+                        {
+                                break;
+                        }
+
                         event.type = keyMappings[i].eventType;
                         break;
                 }
-        }
+        }        
 
         // Handle gg
-        if (event.key == 'g' && event.type == EVENT_NONE)
+        if (event.key[0] == 'g' && event.type == EVENT_NONE)
         {
                 if (gPressed)
                 {
@@ -227,10 +253,10 @@ struct Event processInput()
         }
 
         // Handle numbers
-        if (isdigit(event.key))
+        if (isdigit(event.key[0]))
         {
                 if (digitsPressedCount < maxDigitsPressedCount)
-                        digitsPressed[digitsPressedCount++] = event.key;
+                        digitsPressed[digitsPressedCount++] = event.key[0];
         }
         else
         {
@@ -289,7 +315,7 @@ struct Event processInput()
         }
 
         // Forget g pressed
-        if (event.key != 'g')
+        if (event.key[0] != 'g')
         {
                 gPressed = false;
         }
@@ -342,78 +368,41 @@ void notifySongSwitch()
         bool isDeleted = (audioData.currentFileIndex == 0) ? userData.songdataADeleted == true : userData.songdataBDeleted == true;
 
         if (isDeleted == false && currentSongData != NULL && currentSongData->hasErrors == 0 && currentSongData->metadata && strlen(currentSongData->metadata->title) > 0)
+        {
                 displaySongNotification(currentSongData->metadata->artist, currentSongData->metadata->title, currentSongData->coverArtPath);
+
+                gint64 length = getLengthInSec(currentSongData->duration);
+                // update mpris
+                emitMetadataChanged(
+                    currentSongData->metadata->title,
+                    currentSongData->metadata->artist,
+                    currentSongData->metadata->album,
+                    currentSongData->coverArtPath,
+                    currentSongData->trackId != NULL ? currentSongData->trackId : "", currentSong,
+                    length);
+        }
 }
 
-void prepareNextSong()
+// Checks conditions for refreshing player
+bool shouldRefreshPlayer()
 {
-        if (!skipOutOfOrder && !isRepeatEnabled())
-        {
-                if (currentSong != NULL)
-                        lastPlayedId = currentSong->id;
-                currentSong = getNextSong();
-        }
-        else
-        {
-                skipOutOfOrder = false;
-        }
-
-        finishLoading();
-        resetTimeCount();
-
-        nextSong = NULL;
-        refresh = true;
-
-        if (!isRepeatEnabled() || currentSong == NULL)
-        {
-                unloadPreviousSong();
-        }
-
-        if (currentSong == NULL)
-        {
-                if (quitAfterStopping)
-                        quit();
-                else
-                        setEndOfListReached();
-        }
-        else
-        {
-                notifySongSwitch();
-        }
-
-        clock_gettime(CLOCK_MONOTONIC, &start_time);
+        return !skipping && !isEOFReached() && !isImplSwitchReached();
 }
 
+// Refreshes the player visually if conditions are met
 void refreshPlayer()
 {
-        if (pthread_mutex_trylock(&switchMutex) != 0)
+        int mutexResult = pthread_mutex_trylock(&switchMutex);
+
+        if (mutexResult != 0)
         {
+                fprintf(stderr, "Failed to lock switch mutex.\n");
                 return;
         }
 
-        SongData *songData = (audioData.currentFileIndex == 0) ? userData.songdataA : userData.songdataB;
-        bool isDeleted = (audioData.currentFileIndex == 0) ? userData.songdataADeleted == true : userData.songdataBDeleted == true;
-
-        if (!skipping && !isEOFReached() && !isImplSwitchReached())
+        if (shouldRefreshPlayer())
         {
-                if (refresh && songData != NULL && isDeleted == false &&
-                    songData->hasErrors == false && currentSong != NULL && songData->metadata != NULL)
-                {
-                        gint64 length = llround(songData->duration * G_USEC_PER_SEC);
-                        // update mpris
-                        emitMetadataChanged(
-                            songData->metadata->title,
-                            songData->metadata->artist,
-                            songData->metadata->album,
-                            songData->coverArtPath,
-                            songData->trackId != NULL ? songData->trackId : "", currentSong,
-                            length);
-                }
-
-                if (isDeleted || songData->hasErrors)
-                        songData = NULL;
-
-                printPlayer(songData, elapsedSeconds, &settings);
+                printPlayer(getCurrentSongData(), elapsedSeconds, &settings);
         }
 
         pthread_mutex_unlock(&switchMutex);
@@ -426,7 +415,42 @@ void handleGoToSong()
 
         goingToSong = true;
 
-        if (appState.currentView != LIBRARY_VIEW)
+        if (appState.currentView == LIBRARY_VIEW)
+        {
+                if (audioData.restart)
+                {
+                        Node *lastSong = findSelectedEntryById(&playlist, lastPlayedId);
+                        startFromTop = false;
+
+                        if (lastSong == NULL)
+                        {
+                                if (playlist.tail != NULL)
+                                        lastPlayedId = playlist.tail->id;
+                                else
+                                {
+                                        lastPlayedId = -1;
+                                        startFromTop = true;
+                                }
+                        }
+                }
+
+                pthread_mutex_lock(&(playlist.mutex));
+
+                enqueueSongs(getCurrentLibEntry());
+
+                pthread_mutex_unlock(&(playlist.mutex));
+        }
+        else if (appState.currentView == SEARCH_VIEW)
+        {
+                pthread_mutex_lock(&(playlist.mutex));
+
+                setChosenDir(getCurrentSearchEntry());
+
+                enqueueSongs(getCurrentSearchEntry());
+
+                pthread_mutex_unlock(&(playlist.mutex));                
+        }
+        else
         {
                 if (digitsPressedCount == 0)
                 {
@@ -451,31 +475,6 @@ void handleGoToSong()
                         skipToNumberedSong(songNumber);
                 }
         }
-        else
-        {
-                if (audioData.restart)
-                {
-                        Node *lastSong = findSelectedEntryById(&playlist, lastPlayedId);
-                        startFromTop = false;
-
-                        if (lastSong == NULL)
-                        {
-                                if (playlist.tail != NULL)
-                                        lastPlayedId = playlist.tail->id;
-                                else
-                                {
-                                        lastPlayedId = -1;
-                                        startFromTop = true;
-                                }
-                        }
-                }
-
-                pthread_mutex_lock(&(playlist.mutex));
-
-                enqueueSongs();
-
-                pthread_mutex_unlock(&(playlist.mutex));
-        }
 
         goingToSong = false;
 }
@@ -580,6 +579,9 @@ void handleInput()
         case EVENT_SHOWINFO:
                 toggleShowPlaylist();
                 break;
+        case EVENT_SHOWSEARCH:
+                toggleShowSearch();
+                break;
         case EVENT_SHOWLIBRARY:
                 toggleShowLibrary();
                 break;
@@ -685,7 +687,7 @@ void loadAudioData()
 
                         if (res >= 0)
                         {
-                                resumePlayback();                            
+                                resumePlayback();
                         }
                         else
                         {
@@ -697,7 +699,7 @@ void loadAudioData()
                         nextSong = NULL;
                         refresh = true;
 
-                        clock_gettime(CLOCK_MONOTONIC, &start_time);                        
+                        clock_gettime(CLOCK_MONOTONIC, &start_time);
                 }
         }
         else if (currentSong != NULL && (nextSongNeedsRebuilding || nextSong == NULL) && !songLoading)
@@ -737,10 +739,65 @@ void tryLoadNext()
         }
 }
 
+void setCurrentSongToNext()
+{
+        if (currentSong != NULL)
+                lastPlayedId = currentSong->id;
+        currentSong = getNextSong();
+}
+
+void prepareNextSong()
+{
+        if (!skipOutOfOrder && !isRepeatEnabled())
+        {
+                setCurrentSongToNext();
+        }
+        else
+        {
+                skipOutOfOrder = false;
+        }
+
+        finishLoading();
+        resetTimeCount();
+
+        nextSong = NULL;
+        refresh = true;
+
+        if (!isRepeatEnabled() || currentSong == NULL)
+        {
+                unloadPreviousSong();
+        }
+
+        if (currentSong == NULL)
+        {
+                if (quitAfterStopping)
+                        quit();
+                else
+                        setEndOfListReached();
+        }
+        else
+        {
+                notifySongSwitch();
+        }
+
+        clock_gettime(CLOCK_MONOTONIC, &start_time);
+}
+
+void handleSkipFromStopped()
+{
+        // If we don't do this the song gets loaded in the wrong slot
+        if (skipFromStopped)
+        {
+                usingSongDataA = !usingSongDataA;
+                skipOutOfOrder = false;
+                skipFromStopped = false;
+        }
+}
+
 gboolean mainloop_callback(gpointer data)
 {
         (void)data;
-        
+
         calcElapsedTime();
         handleInput();
 
@@ -750,12 +807,17 @@ gboolean mainloop_callback(gpointer data)
                 g_main_context_iteration(global_main_context, FALSE);
         }
 
-        updatePlayer();
+        updateCounter++;
+        if (updateCounter % 2 == 0 || appState.currentView != SONG_VIEW) // Update every other time or if searching
+        {
+                updatePlayer();
+        }
 
         if (playlist.head != NULL)
         {
-                if (loadingAudioData == false && (!loadedNextSong || nextSongNeedsRebuilding) && !audioData.endOfListReached)
+                if (loadingAudioData == false && (skipFromStopped || !loadedNextSong || nextSongNeedsRebuilding) && !audioData.endOfListReached)
                 {
+                        handleSkipFromStopped();
                         loadAudioData();
                 }
 
@@ -766,6 +828,7 @@ gboolean mainloop_callback(gpointer data)
                 {
                         updateLastSongSwitchTime();
                         prepareNextSong();
+
                         if (!doQuit)
                                 switchAudioImplementation();
                 }
@@ -809,7 +872,7 @@ void play(Node *song)
                 if (res < 0)
                         setEndOfListReached();
         }
-                
+
         if (song == NULL || res < 0)
         {
                 song = NULL;
@@ -831,7 +894,7 @@ void play(Node *song)
         else
                 emitPlaybackStoppedMpris();
 
-        g_timeout_add(100, mainloop_callback, NULL);
+        g_timeout_add(50, mainloop_callback, NULL);
         g_main_loop_run(main_loop);
         g_main_loop_unref(main_loop);
 }
@@ -852,7 +915,7 @@ void cleanupOnExit()
         {
                 printf("No Music found.\n");
                 printf("Please make sure the path is set correctly. \n");
-                printf("To set it type: kew path \"/path/to/Music\". \n");                
+                printf("To set it type: kew path \"/path/to/Music\". \n");
         }
 
         if (!userData.songdataADeleted)
@@ -866,11 +929,12 @@ void cleanupOnExit()
                 unloadSongData(&loadingdata.songdataB);
         }
 
+        freeSearchResults();
         cleanupMpris();
         restoreTerminalMode();
         enableInputBuffering();
         setConfig(&settings);
-        saveSpecialPlaylist(settings.path, playingMainPlaylist);
+        saveSpecialPlaylist(settings.path);
         freeAudioBuffer();
         deleteCache(tempCache);
         deleteTempDir();
@@ -919,7 +983,7 @@ void run()
 }
 
 void init()
-{       
+{
         disableInputBuffering();
         srand(time(NULL));
         initResize();
@@ -942,8 +1006,9 @@ void init()
         pthread_mutex_init(&(loadingdata.mutex), NULL);
         pthread_mutex_init(&(playlist.mutex), NULL);
         nerdFontsEnabled = hasNerdFonts();
-        fflush(stdout);
         createLibrary(&settings);
+        setlocale(LC_ALL, "");
+        fflush(stdout);
 
 #ifdef DEBUG
         g_setenv("G_MESSAGES_DEBUG", "all", TRUE);
@@ -973,7 +1038,7 @@ void playSpecialPlaylist()
                 printf("Couldn't find any songs in the special playlist. Add a song by pressing '.' while it's playing. \n");
                 exit(0);
         }
-        
+
         playingMainPlaylist = true;
 
         init();
@@ -1083,6 +1148,7 @@ int main(int argc, char *argv[])
                 exit(0);
         }
 
+        moveConfigFiles(); // move from old location to new
         getConfig(&settings);
 
         if (argc == 3 && (strcmp(argv[1], "path") == 0))
