@@ -11,16 +11,21 @@ playerops.c
 #define MAXPATHLEN 4096
 #endif
 
+#ifndef ASK_IF_USE_CACHE_LIMIT_SECONDS
+#define ASK_IF_USE_CACHE_LIMIT_SECONDS 4
+#endif
+
 struct timespec current_time;
 struct timespec start_time;
 struct timespec pause_time;
 struct timespec lastInputTime;
 struct timespec lastPlaylistChangeTime;
+struct timespec lastUpdateTime = {0, 0};
 
 bool playlistNeedsUpdate = false;
 bool nextSongNeedsRebuilding = false;
 bool enqueuedNeedsUpdate = false;
-
+bool skipFromStopped = false;
 bool playingMainPlaylist = false;
 bool usingSongDataA = true;
 
@@ -35,7 +40,7 @@ Node *prevSong = NULL;
 int lastPlayedId = -1;
 
 bool songHasErrors = false;
-bool skipPrev = false;
+bool skipOutOfOrder = false;
 bool skipping = false;
 bool loadingFailed = false;
 bool forceSkip = false;
@@ -49,21 +54,15 @@ void updatePlaylist()
 
         if (isShuffleEnabled())
         {
-                shufflePlaylistStartingFromSong(&playlist, currentSong);
+                if (currentSong != NULL)
+                        shufflePlaylistStartingFromSong(&playlist, currentSong);
+                else
+                        shufflePlaylist(&playlist);
 
                 nextSongNeedsRebuilding = true;
         }
 }
 
-void updateNextSong()
-{
-        nextSongNeedsRebuilding = false;
-        nextSong = NULL;
-        nextSong = getNextSong();
-        rebuildNextSong(nextSong);
-        loadedNextSong = true;
-}
-
 void rebuildAndUpdatePlaylist()
 {
         if (!playlistNeedsUpdate && !nextSongNeedsRebuilding)
@@ -76,33 +75,30 @@ void rebuildAndUpdatePlaylist()
                 updatePlaylist();
         }
 
-        if (nextSongNeedsRebuilding)
-        {
-                updateNextSong();
-        }
-
         pthread_mutex_unlock(&(playlist.mutex));
 }
 
 void skip()
 {
         setCurrentImplementationType(NONE);
-        
+
         setRepeatEnabled(false);
         audioData.endOfListReached = false;
 
         rebuildAndUpdatePlaylist();
 
         if (!isPlaying())
-        {                            
+        {
                 switchAudioImplementation();
+                skipFromStopped = true;
         }
         else
         {
                 setSkipToNext(true);
         }
-        
-        refresh = true;
+
+        if (!skipOutOfOrder)
+                refresh = true;
 }
 
 void updateLastSongSwitchTime()
@@ -114,12 +110,43 @@ void updateLastPlaylistChangeTime()
 {
         clock_gettime(CLOCK_MONOTONIC, &lastPlaylistChangeTime);
 }
-
 void updateLastInputTime()
 {
         clock_gettime(CLOCK_MONOTONIC, &lastInputTime);
 }
 
+void updatePlaybackPosition(double elapsedSeconds)
+{
+        GVariantBuilder changedPropertiesBuilder;
+        g_variant_builder_init(&changedPropertiesBuilder, G_VARIANT_TYPE_DICTIONARY);
+        g_variant_builder_add(&changedPropertiesBuilder, "{sv}", "Position", g_variant_new_int64(llround(elapsedSeconds * G_USEC_PER_SEC)));
+
+        GVariant *parameters = g_variant_new("(sa{sv}as)", "org.mpris.MediaPlayer2.Player", &changedPropertiesBuilder, NULL);
+
+        g_dbus_connection_emit_signal(connection,
+                                      NULL,
+                                      "/org/mpris/MediaPlayer2",
+                                      "org.freedesktop.DBus.Properties",
+                                      "PropertiesChanged",
+                                      parameters,
+                                      NULL);
+}
+
+void emitSeekedSignal(double newPositionSeconds)
+{
+        gint64 newPositionMicroseconds = llround(newPositionSeconds * G_USEC_PER_SEC);
+
+        GVariant *parameters = g_variant_new("(x)", newPositionMicroseconds);
+
+        g_dbus_connection_emit_signal(connection,
+                                      NULL,
+                                      "/org/mpris/MediaPlayer2",
+                                      "org.mpris.MediaPlayer2.Player",
+                                      "Seeked",
+                                      parameters,
+                                      NULL);
+}
+
 void emitStringPropertyChanged(const gchar *propertyName, const gchar *newValue)
 {
         GVariantBuilder changed_properties_builder;
@@ -140,7 +167,7 @@ void emitBooleanPropertyChanged(const gchar *propertyName, gboolean newValue)
         g_variant_builder_clear(&changed_properties_builder);
 }
 
-void playbackPause(double *totalPauseSeconds, double *pauseSeconds, struct timespec *pause_time)
+void playbackPause(struct timespec *pause_time)
 {
         if (!isPaused())
         {
@@ -151,7 +178,7 @@ void playbackPause(double *totalPauseSeconds, double *pauseSeconds, struct times
         pausePlayback();
 }
 
-void playbackPlay(double *totalPauseSeconds, double *pauseSeconds, struct timespec *pause_time)
+void playbackPlay(double *totalPauseSeconds, double *pauseSeconds)
 {
         if (isPaused())
         {
@@ -195,15 +222,20 @@ void toggleRepeat()
                 refresh = true;
 }
 
-void addToPlaylist()
+void addToSpecialPlaylist()
 {
-        if (!playingMainPlaylist)
-        {
-                SongInfo song;
-                song.filePath = strdup(currentSong->song.filePath);
-                song.duration = 0.0;
-                addToList(mainPlaylist, song, nodeIdCounter++);
-        }
+        if (currentSong == NULL)
+                return;
+
+        int id = currentSong->id;
+
+        Node *node = NULL;
+
+        if (findSelectedEntryById(specialPlaylist, id) != NULL) // song is already in list
+                return;
+
+        createNode(&node, currentSong->song.filePath, id);
+        addToList(specialPlaylist, node);
 }
 
 void toggleShuffle()
@@ -213,13 +245,35 @@ void toggleShuffle()
 
         if (shuffleEnabled)
         {
+                pthread_mutex_lock(&(playlist.mutex));
+
                 shufflePlaylistStartingFromSong(&playlist, currentSong);
+
+                pthread_mutex_unlock(&(playlist.mutex));
+
                 emitBooleanPropertyChanged("Shuffle", TRUE);
         }
         else
         {
-                playlist = deepCopyPlayList(originalPlaylist);                
-                currentSong = findSongInPlaylist(currentSong, &playlist);
+                char *path = NULL;
+                if (currentSong != NULL)
+                {
+                        path = strdup(currentSong->song.filePath);
+                }
+
+                pthread_mutex_lock(&(playlist.mutex));
+
+                deletePlaylist(&playlist); // Doesn't destroy the mutex
+                deepCopyPlayListOntoList(originalPlaylist, &playlist);
+
+                if (path != NULL)
+                {
+                        currentSong = findPathInPlaylist(path, &playlist);
+                        free(path);
+                }
+
+                pthread_mutex_unlock(&(playlist.mutex));
+
                 emitBooleanPropertyChanged("Shuffle", FALSE);
         }
 
@@ -230,10 +284,10 @@ void toggleShuffle()
                 refresh = true;
 }
 
-void toggleBlocks()
+void toggleBlocks(AppSettings *settings)
 {
         coverAnsi = !coverAnsi;
-        c_strcpy(settings.coverAnsi, sizeof(settings.coverAnsi), coverAnsi ? "1" : "0");
+        c_strcpy(settings->coverAnsi, sizeof(settings->coverAnsi), coverAnsi ? "1" : "0");
         if (coverEnabled)
         {
                 clearScreen();
@@ -241,26 +295,26 @@ void toggleBlocks()
         }
 }
 
-void toggleColors()
+void toggleColors(AppSettings *settings)
 {
         useProfileColors = !useProfileColors;
-        c_strcpy(settings.useProfileColors, sizeof(settings.useProfileColors), useProfileColors ? "1" : "0");
+        c_strcpy(settings->useProfileColors, sizeof(settings->useProfileColors), useProfileColors ? "1" : "0");
         clearScreen();
         refresh = true;
 }
 
-void toggleCovers()
+void toggleCovers(AppSettings *settings)
 {
         coverEnabled = !coverEnabled;
-        c_strcpy(settings.coverEnabled, sizeof(settings.coverEnabled), coverEnabled ? "1" : "0");
+        c_strcpy(settings->coverEnabled, sizeof(settings->coverEnabled), coverEnabled ? "1" : "0");
         clearScreen();
         refresh = true;
 }
 
-void toggleVisualizer()
+void toggleVisualizer(AppSettings *settings)
 {
         visualizerEnabled = !visualizerEnabled;
-        c_strcpy(settings.visualizerEnabled, sizeof(settings.visualizerEnabled), visualizerEnabled ? "1" : "0");
+        c_strcpy(settings->visualizerEnabled, sizeof(settings->visualizerEnabled), visualizerEnabled ? "1" : "0");
         restoreCursorPosition();
         refresh = true;
 }
@@ -270,27 +324,46 @@ void quit()
         doQuit = true;
 }
 
-SongData *getCurrentSongData()
+bool isCurrentSongDeleted()
+{
+        return (audioData.currentFileIndex == 0) ? userData.songdataADeleted == true : userData.songdataBDeleted == true;
+}
+
+bool isValidSong(SongData *songData)
+{
+        return songData != NULL && songData->hasErrors == false && songData->metadata != NULL;
+}
+
+SongData *getCurrentSongData(void)
 {
         if (currentSong == NULL)
                 return NULL;
 
-        if (usingSongDataA)
-                return loadingdata.songdataA;
-        else
-                return loadingdata.songdataB;
+        if (isCurrentSongDeleted())
+                return NULL;
+
+        SongData *songData = (audioData.currentFileIndex == 0) ? userData.songdataA : userData.songdataB;
+
+        if (!isValidSong(songData))
+                return NULL;                
+
+        return songData;
 }
 
 void calcElapsedTime()
 {
         clock_gettime(CLOCK_MONOTONIC, &current_time);
+
+        double timeSinceLastUpdate = (double)(current_time.tv_sec - lastUpdateTime.tv_sec) +
+                                     (double)(current_time.tv_nsec - lastUpdateTime.tv_nsec) / 1e9;
+
         if (!isPaused())
         {
                 elapsedSeconds = (double)(current_time.tv_sec - start_time.tv_sec) +
                                  (double)(current_time.tv_nsec - start_time.tv_nsec) / 1e9;
                 double seekElapsed = getSeekElapsed();
                 double diff = elapsedSeconds + (seekElapsed + seekAccumulatedSeconds - totalPauseSeconds);
-                
+
                 if (diff < 0)
                         seekElapsed -= diff;
 
@@ -304,6 +377,12 @@ void calcElapsedTime()
                 {
                         elapsedSeconds = 0.0;
                 }
+
+                if (currentSong != NULL && timeSinceLastUpdate >= 1.0)
+                {
+                        updatePlaybackPosition(elapsedSeconds);
+                        lastUpdateTime = current_time;
+                }
         }
         else
         {
@@ -336,6 +415,8 @@ void flushSeek()
                 }
 
                 seekPercentage(percentage);
+
+                emitSeekedSignal(elapsedSeconds);
         }
 }
 
@@ -386,11 +467,15 @@ Node *findSelectedEntryById(PlayList *playlist, int id)
 
         for (int i = 0; i < playlist->count; i++)
         {
-                if (node->id == id)
+                if (node != NULL && node->id == id)
                 {
                         found = true;
                         break;
                 }
+                else if (node == NULL)
+                {
+                        return NULL;
+                }
                 node = node->next;
         }
 
@@ -479,22 +564,6 @@ bool markAsDequeued(FileSystemEntry *root, char *path)
         return false;
 }
 
-void removeSelectedEntry(PlayList *playlist, Node *node)
-{
-        if (node != NULL)
-                markAsDequeued(getLibrary(), node->song.filePath);
-
-        deleteFromList(playlist, node);
-}
-
-void removeSelectedEntryByPath(PlayList *playlist, char *path)
-{
-        if (path != NULL)
-                markAsDequeued(getLibrary(), path);
-        Node *node = findLastPathInPlaylist(path, playlist);
-        deleteFromList(playlist, node);
-}
-
 Node *getNextSong()
 {
         if (nextSong != NULL)
@@ -511,12 +580,15 @@ Node *getNextSong()
 
 void enqueueSong(FileSystemEntry *child)
 {
-        SongInfo song;
-        song.filePath = strdup(child->fullPath);
-        song.duration = 0.0;
-        nodeIdCounter++;
-        addToList(originalPlaylist, song, nodeIdCounter);
-        addToList(&playlist, song, nodeIdCounter);
+        int id = nodeIdCounter++;
+
+        Node *node = NULL;
+        createNode(&node, child->fullPath, id);
+        addToList(originalPlaylist, node);
+
+        Node *node2 = NULL;
+        createNode(&node2, child->fullPath, id);
+        addToList(&playlist, node2);
 
         child->isEnqueued = 1;
         child->parent->isEnqueued = 1;
@@ -524,7 +596,6 @@ void enqueueSong(FileSystemEntry *child)
 
 void dequeueSong(FileSystemEntry *child)
 {
-        child->isEnqueued = 0;
         Node *node1 = findLastPathInPlaylist(child->fullPath, originalPlaylist);
 
         if (node1 == NULL)
@@ -535,24 +606,48 @@ void dequeueSong(FileSystemEntry *child)
 
         int id = node1->id;
 
+        Node *node2 = findSelectedEntryById(&playlist, id);
+
         if (node1 != NULL)
                 deleteFromList(originalPlaylist, node1);
 
-        Node *node2 = findSelectedEntryById(&playlist, id);
-        deleteFromList(&playlist, node2);
+        if (node2 != NULL)
+                deleteFromList(&playlist, node2);
 
         child->isEnqueued = 0;
-        child->parent->isEnqueued = 0;
+
+        // check if parent needs to be dequeued as well
+        bool isEnqueued = false;
+
+        FileSystemEntry *ch = child->parent->children;
+
+        while (ch != NULL)
+        {
+                if (ch->isEnqueued)
+                {
+                        isEnqueued = true;
+                        break;
+                }
+                ch = ch->next;
+        }
+
+        if (!isEnqueued)
+        {
+                child->parent->isEnqueued = 0;
+                if (child->parent->parent != NULL)
+                        child->parent->parent->isEnqueued = 0;
+        }
 }
 
-void dequeueChildren(FileSystemEntry *child)
+void dequeueChildren(FileSystemEntry *parent)
 {
+        FileSystemEntry *child = parent->children;
+
         while (child != NULL)
         {
                 if (child->isDirectory && child->children != NULL)
                 {
-                        child->isEnqueued = 0;
-                        dequeueChildren(child->children);
+                        dequeueChildren(child);
                 }
                 else
                 {
@@ -572,7 +667,7 @@ void enqueueChildren(FileSystemEntry *child)
                         child->isEnqueued = 1;
                         enqueueChildren(child->children);
                 }
-                else
+                else if (!child->isEnqueued)
                 {
                         enqueueSong(child);
                 }
@@ -602,24 +697,45 @@ bool hasSongChildren(FileSystemEntry *entry)
         return true;
 }
 
-void enqueueSongs()
+bool hasDequeuedChildren(FileSystemEntry *parent)
+{
+        FileSystemEntry *child = parent->children;
+
+        bool isDequeued = false;
+        while (child != NULL)
+        {
+                if (!child->isEnqueued)
+                {
+                        isDequeued = true;
+                }
+                child = child->next;
+        }
+
+        return isDequeued;
+}
+
+void dequeueAll()
+{
+        
+}
+
+void enqueueSongs(FileSystemEntry *entry)
 {
-        FileSystemEntry *tmp = getCurrentLibEntry();
         FileSystemEntry *chosenDir = getChosenDir();
         bool hasEnqueued = false;
 
-        if (tmp != NULL)
+        if (entry != NULL)
         {
-                if (tmp->isDirectory)
+                if (entry->isDirectory)
                 {
-                        if (!hasSongChildren(tmp) || (chosenDir != NULL && strcmp(tmp->fullPath, chosenDir->fullPath) == 0))
+                        if (!hasSongChildren(entry) || (chosenDir != NULL && strcmp(entry->fullPath, chosenDir->fullPath) == 0))
                         {
-                                if (!tmp->isEnqueued)
+                                if (hasDequeuedChildren(entry))
                                 {
-                                        tmp->isEnqueued = 1;
-                                        tmp = tmp->children;
+                                        entry->isEnqueued = 1;
+                                        entry = entry->children;
 
-                                        enqueueChildren(tmp);
+                                        enqueueChildren(entry);
 
                                         nextSongNeedsRebuilding = true;
 
@@ -627,10 +743,7 @@ void enqueueSongs()
                                 }
                                 else
                                 {
-                                        tmp->isEnqueued = 0;
-                                        tmp = tmp->children;
-
-                                        dequeueChildren(tmp);
+                                        dequeueChildren(entry);
 
                                         nextSongNeedsRebuilding = true;
                                 }
@@ -641,27 +754,21 @@ void enqueueSongs()
                 }
                 else
                 {
-                        if (!tmp->isEnqueued)
+                        if (!entry->isEnqueued)
                         {
-                                Node *song = getNextSong();
-                                if (song == NULL || strcmp(tmp->fullPath, song->song.filePath) == 0)
-                                {
-                                        nextSong = NULL;
-                                        nextSongNeedsRebuilding = true;
-                                }
-                                enqueueSong(tmp);
+                                nextSong = NULL;
+                                nextSongNeedsRebuilding = true;
+
+                                enqueueSong(entry);
 
                                 hasEnqueued = true;
                         }
                         else
                         {
-                                Node *song = getNextSong();
-                                if (song == NULL || strcmp(tmp->fullPath, song->song.filePath) == 0)
-                                {
-                                        nextSong = NULL;
-                                        nextSongNeedsRebuilding = true;
-                                }
-                                dequeueSong(tmp);
+                                nextSong = NULL;
+                                nextSongNeedsRebuilding = true;
+
+                                dequeueSong(entry);
                         }
                 }
                 refresh = true;
@@ -670,77 +777,89 @@ void enqueueSongs()
         if (hasEnqueued)
         {
                 waitingForNext = true;
-                loadedNextSong = false;
                 audioData.endOfListReached = false;
         }
 
-        playlistNeedsUpdate = true;
+        if (nextSongNeedsRebuilding)
+        {
+                updatePlaylist();
+        }
         updateLastPlaylistChangeTime();
 }
 
-void resetList()
-{
-        audioData.currentFileIndex = 1 - audioData.currentFileIndex;
-        setCurrentImplementationType(NONE);
-}
-
 void handleRemove()
 {
         if (refresh)
                 return;
 
-        pthread_mutex_lock(&(playlist.mutex));
-
-        bool rebuild = false;
-
         if (appState.currentView != PLAYLIST_VIEW)
         {
-                pthread_mutex_unlock(&(playlist.mutex));
                 return;
         }
 
+        bool rebuild = false;
+
         Node *node = findSelectedEntry(originalPlaylist, chosenRow);
+
+        if (node == NULL)
+        {
+                return;
+        }
+
         Node *song = getNextSong();
         int id = node->id;
+
         int currentId = (currentSong != NULL) ? currentSong->id : -1;
 
         if (node != NULL && playlist.head != NULL && playlist.head->id == node->id && playlist.count == 1)
         {
-                pthread_mutex_unlock(&(playlist.mutex));
                 return;
         }
 
         if (currentId == node->id)
         {
-                pthread_mutex_unlock(&(playlist.mutex));
                 return;
         }
 
+        pthread_mutex_lock(&(playlist.mutex));
+
         if (node != NULL && song != NULL)
         {
-                if (strcmp(song->song.filePath, node->song.filePath) == 0)
+                if (strcmp(song->song.filePath, node->song.filePath) == 0 || (currentSong != NULL && currentSong->next != NULL && id == currentSong->next->id))
                         rebuild = true;
         }
 
-        removeSelectedEntry(originalPlaylist, node);
+        if (node != NULL)
+                markAsDequeued(getLibrary(), node->song.filePath);
 
         Node *node2 = findSelectedEntryById(&playlist, id);
-        deleteFromList(&playlist, node2);
 
-        playlistNeedsUpdate = true;
+        if (node != NULL)
+                deleteFromList(originalPlaylist, node);
+
+        if (node2 != NULL)
+                deleteFromList(&playlist, node2);
+
         updateLastPlaylistChangeTime();
 
         if (isShuffleEnabled())
                 rebuild = true;
 
+        currentSong = findSelectedEntryById(&playlist, currentId);
+
         if (rebuild)
         {
                 node = NULL;
                 nextSong = NULL;
-                nextSongNeedsRebuilding = true;
-        }
+                updatePlaylist();
 
-        currentSong = findSelectedEntryById(&playlist, currentId);
+                tryNextSong = currentSong->next;
+                nextSongNeedsRebuilding = false;
+                nextSong = NULL;
+                nextSong = getListNext(currentSong);
+                rebuildNextSong(nextSong);
+                loadedNextSong = true;
+        }
 
         pthread_mutex_unlock(&(playlist.mutex));
 
@@ -770,44 +889,44 @@ Node *getSongByNumber(PlayList *playlist, int songNumber)
         return song;
 }
 
+int loadDecoder(SongData *songData, bool *songDataDeleted)
+{
+        int result = 0;
+        if (songData != NULL)
+        {
+                *songDataDeleted = false;
+
+                // this should only be done for the second song, as switchAudioImplementation() handles the first one
+                if (!loadingdata.loadingFirstDecoder)
+                {
+                        if (hasBuiltinDecoder(songData->filePath))
+                                result = prepareNextDecoder(songData->filePath);
+                        else if (endsWith(songData->filePath, "opus"))
+                                result = prepareNextOpusDecoder(songData->filePath);
+                        else if (endsWith(songData->filePath, "ogg"))
+                                result = prepareNextVorbisDecoder(songData->filePath);
+                        else if (endsWith(songData->filePath, "m4a") || endsWith(songData->filePath, "aac") || endsWith(songData->filePath, "mp4"))
+                                result = prepareNextM4aDecoder(songData->filePath);
+                }
+        }
+        return result;
+}
+
 int assignLoadedData()
 {
         int result = 0;
 
         if (loadingdata.loadA)
         {
-                if (loadingdata.songdataA != NULL)
-                {                        
-                        userData.filenameA = loadingdata.songdataA->pcmFilePath;
-                        userData.songdataA = loadingdata.songdataA;
-
-                        if (hasBuiltinDecoder(loadingdata.songdataA->filePath))
-                                result = prepareNextDecoder(loadingdata.songdataA->filePath);
-                        else if (endsWith(loadingdata.songdataA->filePath, "opus"))
-                                result = prepareNextOpusDecoder(loadingdata.songdataA->filePath);
-                        else if (endsWith(loadingdata.songdataA->filePath, "ogg"))
-                                result = prepareNextVorbisDecoder(loadingdata.songdataA->filePath);
-                }
-                else
-                        userData.filenameA = NULL;
+                userData.songdataA = loadingdata.songdataA;
+                result = loadDecoder(loadingdata.songdataA, &userData.songdataADeleted);
         }
         else
         {
-                if (loadingdata.songdataB != NULL)
-                {
-                        userData.filenameB = loadingdata.songdataB->pcmFilePath;
-                        userData.songdataB = loadingdata.songdataB;
-
-                        if (hasBuiltinDecoder(loadingdata.songdataB->filePath))
-                                result = prepareNextDecoder(loadingdata.songdataB->filePath);
-                        else if (endsWith(loadingdata.songdataB->filePath, "opus"))
-                                result = prepareNextOpusDecoder(loadingdata.songdataB->filePath);
-                        else if (endsWith(loadingdata.songdataB->filePath, "ogg"))
-                                result = prepareNextVorbisDecoder(loadingdata.songdataB->filePath);
-                }
-                else
-                        userData.filenameB = NULL;
+                userData.songdataB = loadingdata.songdataB;
+                result = loadDecoder(loadingdata.songdataB, &userData.songdataBDeleted);
         }
+
         return result;
 }
 
@@ -820,16 +939,24 @@ void *songDataReaderThread(void *arg)
 
         char filepath[MAXPATHLEN];
         c_strcpy(filepath, sizeof(filepath), loadingdata->filePath);
-        
+
         SongData *songdata = NULL;
 
         if (loadingdata->loadA)
         {
-                unloadSongData(&loadingdata->songdataA);
+                if (!userData.songdataADeleted)
+                {
+                        userData.songdataADeleted = true;
+                        unloadSongData(&loadingdata->songdataA);
+                }
         }
         else
         {
-                unloadSongData(&loadingdata->songdataB);
+                if (!userData.songdataBDeleted)
+                {
+                        userData.songdataBDeleted = true;
+                        unloadSongData(&loadingdata->songdataB);
+                }
         }
 
         if (filepath[0] != '\0')
@@ -841,7 +968,7 @@ void *songDataReaderThread(void *arg)
 
         if (loadingdata->loadA)
         {
-                loadingdata->songdataA = songdata;                
+                loadingdata->songdataA = songdata;
         }
         else
         {
@@ -864,6 +991,7 @@ void *songDataReaderThread(void *arg)
         }
         else
         {
+                songHasErrors = false;
                 clearingErrors = false;
                 nextSong = tryNextSong;
                 tryNextSong = NULL;
@@ -915,15 +1043,12 @@ void rebuildNextSong(Node *song)
                 return;
 
         loadingdata.loadA = !usingSongDataA;
-
-        pthread_mutex_lock(&(loadingdata.mutex));
+        loadingdata.loadingFirstDecoder = false;
 
         songLoading = true;
 
         loadSong(song, &loadingdata);
 
-        pthread_mutex_unlock(&(loadingdata.mutex));
-
         int maxNumTries = 50;
         int numtries = 0;
 
@@ -932,6 +1057,7 @@ void rebuildNextSong(Node *song)
                 c_sleep(100);
                 numtries++;
         }
+        songLoading = false;
 }
 
 void skipToNextSong()
@@ -941,39 +1067,49 @@ void skipToNextSong()
                 return;
         }
 
-        if (songLoading || !loadedNextSong || skipping || clearingErrors)
+        if (songLoading || nextSongNeedsRebuilding || skipping || clearingErrors)
                 return;
 
-        playbackPlay(&totalPauseSeconds, &pauseSeconds, &pause_time);
+        playbackPlay(&totalPauseSeconds, &pauseSeconds);
 
         skipping = true;
+
         updateLastSongSwitchTime();
+
         skip();
 }
 
 void skipToPrevSong()
 {
-        if (currentSong == NULL || currentSong->prev == NULL)
+        if ((currentSong == NULL || currentSong->prev == NULL) && !isShuffleEnabled())
         {
                 return;
         }
 
-        if (songLoading || !loadedNextSong || skipping || clearingErrors)
+        if (songLoading || skipping || clearingErrors)
                 if (!forceSkip)
                         return;
 
-        currentSong = currentSong->prev;
+        if (isShuffleEnabled() && currentSong != NULL && currentSong->prev == NULL)
+        {
+                if (currentSong->prev == NULL && currentSong->next != NULL)
+                        currentSong = currentSong->next;
+                else
+                        return;
+        }
+        else
+                currentSong = currentSong->prev;
 
-        playbackPlay(&totalPauseSeconds, &pauseSeconds, &pause_time);
+        playbackPlay(&totalPauseSeconds, &pauseSeconds);
 
         skipping = true;
-        skipPrev = true;
+        skipOutOfOrder = true;
         loadedNextSong = false;
         songLoading = true;
         forceSkip = false;
 
         loadingdata.loadA = !usingSongDataA;
-
+        loadingdata.loadingFirstDecoder = true;
         loadSong(currentSong, &loadingdata);
         int maxNumTries = 50;
         int numtries = 0;
@@ -992,34 +1128,35 @@ void skipToPrevSong()
         }
 
         updateLastSongSwitchTime();
-
         skip();
 }
 
 void skipToSong(int id)
 {
-        if (songLoading || !loadedNextSong || skipping || clearingErrors)
+        if (songLoading || !loadedNextSong || skipping || skipOutOfOrder || clearingErrors)
                 if (!forceSkip)
                         return;
 
-        playbackPlay(&totalPauseSeconds, &pauseSeconds, &pause_time);
-
-        skipping = true;
-        skipPrev = true;
-        loadedNextSong = false;
-        songLoading = true;
-        forceSkip = false;
-
         Node *found = NULL;
         findNodeInList(&playlist, id, &found);
 
         if (found != NULL)
                 currentSong = found;
         else
+        {
                 return;
+        }
+
+        skipping = true;
+        skipOutOfOrder = true;
+        loadedNextSong = false;
+        songLoading = true;
+        forceSkip = false;
+
+        playbackPlay(&totalPauseSeconds, &pauseSeconds);
 
         loadingdata.loadA = !usingSongDataA;
-
+        loadingdata.loadingFirstDecoder = true;
         loadSong(currentSong, &loadingdata);
         int maxNumTries = 50;
         int numtries = 0;
@@ -1048,10 +1185,10 @@ void skipToNumberedSong(int songNumber)
                 if (!forceSkip)
                         return;
 
-        playbackPlay(&totalPauseSeconds, &pauseSeconds, &pause_time);
+        playbackPlay(&totalPauseSeconds, &pauseSeconds);
 
         skipping = true;
-        skipPrev = true;
+        skipOutOfOrder = true;
         loadedNextSong = false;
         songLoading = true;
         forceSkip = false;
@@ -1059,7 +1196,7 @@ void skipToNumberedSong(int songNumber)
         currentSong = getSongByNumber(originalPlaylist, songNumber);
 
         loadingdata.loadA = !usingSongDataA;
-
+        loadingdata.loadingFirstDecoder = true;
         loadSong(currentSong, &loadingdata);
         int maxNumTries = 50;
         int numtries = 0;
@@ -1102,7 +1239,7 @@ void loadFirstSong(Node *song)
 {
         if (song == NULL)
                 return;
-
+        loadingdata.loadingFirstDecoder = true;
         loadSong(song, &loadingdata);
 
         int i = 0;
@@ -1116,7 +1253,35 @@ void loadFirstSong(Node *song)
         }
 }
 
-void loadFirst(Node *song)
+void unloadPreviousSong()
+{
+        pthread_mutex_lock(&(loadingdata.mutex));
+
+        if (usingSongDataA &&
+            (skipping || (userData.currentSongData == NULL || userData.songdataADeleted == false ||
+                          (loadingdata.songdataA != NULL && userData.songdataADeleted == false && userData.currentSongData->hasErrors == 0 && userData.currentSongData->trackId != NULL && strcmp(loadingdata.songdataA->trackId, userData.currentSongData->trackId) != 0))))
+        {
+                userData.songdataADeleted = true;
+                unloadSongData(&loadingdata.songdataA);
+                usingSongDataA = false;
+                if (!audioData.endOfListReached)
+                        loadedNextSong = false;
+        }
+        else if (!usingSongDataA &&
+                 (skipping || (userData.currentSongData == NULL || userData.songdataBDeleted == false ||
+                               (loadingdata.songdataB != NULL && userData.songdataBDeleted == false && userData.currentSongData->hasErrors == 0 && userData.currentSongData->trackId != NULL && strcmp(loadingdata.songdataB->trackId, userData.currentSongData->trackId) != 0))))
+        {
+                userData.songdataBDeleted = true;
+                unloadSongData(&loadingdata.songdataB);
+                usingSongDataA = true;
+                if (!audioData.endOfListReached)
+                        loadedNextSong = false;
+        }
+
+        pthread_mutex_unlock(&(loadingdata.mutex));
+}
+
+int loadFirst(Node *song)
 {
         loadFirstSong(song);
 
@@ -1132,12 +1297,122 @@ void loadFirst(Node *song)
 
         if (songHasErrors)
         {
-                printf("Couldn't play any of the songs.\n");
-                return;
+                // couldn't play any of the songs
+                unloadPreviousSong();
+                currentSong = NULL;
+                songHasErrors = false;
+                return -1;
         }
 
         userData.currentPCMFrame = 0;
-        //userData.filenameA = loadingdata.songdataA->pcmFilePath;
-        //userData.songdataA = loadingdata.songdataA;
         userData.currentSongData = userData.songdataA;
+
+        return 0;
+}
+
+void *updateLibraryThread(void *arg)
+{
+        char *path = (char *)arg;
+        int tmpDirectoryTreeEntries = 0;
+
+        FileSystemEntry *temp = createDirectoryTree(path, &tmpDirectoryTreeEntries);
+
+        pthread_mutex_lock(&switchMutex);
+
+        freeTree(library);
+        library = temp;
+        numDirectoryTreeEntries = tmpDirectoryTreeEntries;
+        resetChosenDir();
+
+        pthread_mutex_unlock(&switchMutex);
+
+        refresh = true;
+
+        return NULL;
+}
+
+void updateLibrary(char *path)
+{
+        pthread_t threadId;
+
+        if (pthread_create(&threadId, NULL, updateLibraryThread, path) != 0)
+        {
+                perror("Failed to create thread");
+                return;
+        }
+}
+
+void askIfCacheLibrary()
+{
+        if (cacheLibrary > -1) // Only use this function if cacheLibrary isn't set
+                return;
+
+        char input = '\0';
+
+        restoreTerminalMode();
+        enableInputBuffering();
+        showCursor();
+
+        printf("Would you like to enable a (local) library cache for quicker startup times?\nYou can update the cache at any time by pressing 'u'. (y/n): ");
+
+        fflush(stdout);
+
+        do
+        {
+                int res = scanf(" %c", &input);
+
+                if (res < 0)
+                        break;
+
+        } while (input != 'Y' && input != 'y' && input != 'N' && input != 'n');
+
+        if (input == 'Y' || input == 'y')
+        {
+                printf("Y\n");
+                cacheLibrary = 1;
+        }
+        else
+        {
+                printf("N\n");
+                cacheLibrary = 0;
+        }
+
+        setNonblockingMode();
+        disableInputBuffering();
+        hideCursor();
+}
+
+void createLibrary(AppSettings *settings)
+{
+        if (cacheLibrary > 0)
+        {
+                char *libFilepath = getLibraryFilePath();
+                library = reconstructTreeFromFile(libFilepath, settings->path, &numDirectoryTreeEntries);
+                free(libFilepath);
+        }
+
+        if (library == NULL || library->children == NULL)
+        {
+                struct timeval start, end;
+
+                gettimeofday(&start, NULL);
+
+                library = createDirectoryTree(settings->path, &numDirectoryTreeEntries);
+
+                gettimeofday(&end, NULL);
+                long seconds = end.tv_sec - start.tv_sec;
+                long microseconds = end.tv_usec - start.tv_usec;
+                double elapsed = seconds + microseconds * 1e-6;
+
+                // If time to load the library was significant, ask to use cache instead
+                if (elapsed > ASK_IF_USE_CACHE_LIMIT_SECONDS)
+                {
+                        askIfCacheLibrary();
+                }
+        }
+
+        if (library == NULL || library->children == NULL)
+        {
+                exit(0);
+        }
 }
