@@ -235,6 +235,14 @@ void opus_createAudioDevice(UserData *userData, ma_device *device, ma_context *c
         emitStringPropertyChanged("PlaybackStatus", "Playing");
 }
 
+bool validFilePath(char *filePath)
+{
+        if (filePath == NULL || filePath[0] == '\0' || filePath[0] == '\r' || existsFile(filePath) < 0)
+                return false;
+        else
+                return true;
+}
+
 int switchAudioImplementation()
 {
         if (audioData.endOfListReached)
@@ -256,7 +264,7 @@ int switchAudioImplementation()
 
         char *filePath = strdup(userData.currentSongData->filePath);
 
-        if (filePath == NULL || filePath[0] == '\0' || filePath[0] == '\r' || existsFile(filePath) < 0)
+        if (!validFilePath(filePath))
         {
                 free(filePath);
                 setEOFReached();
@@ -466,10 +474,23 @@ int createAudioDevice(UserData *userData)
                 SongData *currentSongData = userData->currentSongData;
 
                 if (currentSongData != NULL && currentSongData->hasErrors == 0 && currentSongData->metadata && strlen(currentSongData->metadata->title) > 0)
+                {
                         displaySongNotification(currentSongData->metadata->artist, currentSongData->metadata->title, currentSongData->coverArtPath);
+
+                        gint64 length = getLengthInSec(currentSongData->duration);
+                        // update mpris
+                        emitMetadataChanged(
+                            currentSongData->metadata->title,
+                            currentSongData->metadata->artist,
+                            currentSongData->metadata->album,
+                            currentSongData->coverArtPath,
+                            currentSongData->trackId != NULL ? currentSongData->trackId : "", currentSong,
+                            length);
+                }
         }
-        else {
-                return -1;                
+        else
+        {
+                return -1;
         }
 
         return 0;
