@@ -1126,33 +1126,53 @@ char *remove_blacklisted_chars(const char *input, const char *blacklist)
         return output;
 }
 
-int displaySongNotification(const char *artist, const char *title, const char *cover)
+gint64 getLengthInSec(double duration)
 {
-        if (!allowNotifications)
-                return 0;
+        return floor(llround(duration * G_USEC_PER_SEC));
+}
 
-        char command[MAXPATHLEN + 1024];
-        char sanitized_cover[MAXPATHLEN];
+int displaySongNotification(const char *artist, const char *title, const char *cover) {
+    if (!allowNotifications)
+        return 0;
 
-        const char *blacklist = "&;`|*~<>^()[]{}$\\\"";
-        char *sanitizedArtist = remove_blacklisted_chars(artist, blacklist);
-        char *sanitizedTitle = remove_blacklisted_chars(title, blacklist);
+    char sanitized_cover[MAXPATHLEN];
 
-        sanitize_filepath(cover, sanitized_cover, sizeof(sanitized_cover));
+    const char *blacklist = "&;`|*~<>^()[]{}$\\\"";
+    char *sanitizedArtist = remove_blacklisted_chars(artist, blacklist);
+    char *sanitizedTitle = remove_blacklisted_chars(title, blacklist);
 
-        if (strlen(artist) <= 0)
-        {
-                snprintf(command, sizeof(command), "notify-send -a \"kew\" \"%s\" --icon \"%s\"", sanitizedTitle, sanitized_cover);
-        }
-        else
-        {
-                snprintf(command, sizeof(command), "notify-send -a \"kew\" \"%s - %s\" --icon \"%s\"", sanitizedArtist, sanitizedTitle, sanitized_cover);
+    sanitize_filepath(cover, sanitized_cover, sizeof(sanitized_cover));
+
+    char message[MAXPATHLEN + 1024];
+    if (strlen(artist) > 0) {
+        snprintf(message, sizeof(message), "%s - %s", sanitizedArtist, sanitizedTitle);
+    } else {
+        snprintf(message, sizeof(message), "%s", sanitizedTitle);
+    }
+
+    char *args[] = {"/usr/bin/notify-send", "-a", "kew", message, "--icon", sanitized_cover, NULL};
+
+    pid_t pid = fork();
+    if (pid == -1) {
+        // Handle error
+        perror("fork");
+    } else if (pid > 0) {
+        // Parent process
+        int status;
+        waitpid(pid, &status, 0);
+    } else {
+        // Child process
+        extern char **environ;
+        if (execve("/usr/bin/notify-send", args, environ) == -1) {
+
+            _Exit(EXIT_FAILURE);
         }
+    }
 
-        free(sanitizedArtist);
-        free(sanitizedTitle);
+    free(sanitizedArtist);
+    free(sanitizedTitle);
 
-        return system(command);
+    return 0;
 }
 
 void executeSwitch(AudioData *pAudioData)
@@ -1222,12 +1242,12 @@ int adjustVolumePercent(int volumeChange)
 {
         int sysVol = getSystemVolume();
 
-        if (sysVol == 0) 
+        if (sysVol == 0)
                 return 0;
 
         int step = 100 / sysVol * 5;
 
-        int relativeVolChange = volumeChange / 5  * step;
+        int relativeVolChange = volumeChange / 5 * step;
 
         soundVolume += relativeVolChange;
 
@@ -1239,15 +1259,15 @@ int adjustVolumePercent(int volumeChange)
 ma_uint64 lastCursor = 0;
 
 void m4a_read_pcm_frames(ma_data_source *pDataSource, void *pFramesOut, ma_uint64 frameCount, ma_uint64 *pFramesRead)
-{        
+{
         m4a_decoder *m4a = (m4a_decoder *)pDataSource;
-        AudioData *pAudioData = (AudioData *)m4a->pReadSeekTellUserData;        
+        AudioData *pAudioData = (AudioData *)m4a->pReadSeekTellUserData;
         ma_uint64 framesRead = 0;
 
         while (framesRead < frameCount)
         {
                 if (doQuit)
-                        return;                        
+                        return;
 
                 if (isImplSwitchReached())
                         return;
@@ -1255,7 +1275,7 @@ void m4a_read_pcm_frames(ma_data_source *pDataSource, void *pFramesOut, ma_uint6
                 if (pthread_mutex_trylock(&dataSourceMutex) != 0)
                 {
                         return;
-                }                        
+                }
 
                 // Check if a file switch is required
                 if (pAudioData->switchFiles)
@@ -1372,7 +1392,7 @@ void opus_read_pcm_frames(ma_data_source *pDataSource, void *pFramesOut, ma_uint
         while (framesRead < frameCount)
         {
                 if (doQuit)
-                        return;                        
+                        return;
 
                 if (isImplSwitchReached())
                         return;
@@ -1380,7 +1400,7 @@ void opus_read_pcm_frames(ma_data_source *pDataSource, void *pFramesOut, ma_uint
                 if (pthread_mutex_trylock(&dataSourceMutex) != 0)
                 {
                         return;
-                }                        
+                }
 
                 // Check if a file switch is required
                 if (pAudioData->switchFiles)
@@ -1498,11 +1518,10 @@ void vorbis_read_pcm_frames(ma_data_source *pDataSource, void *pFramesOut, ma_ui
                 if (isImplSwitchReached())
                         return;
 
-
                 if (pthread_mutex_trylock(&dataSourceMutex) != 0)
                 {
                         return;
-                }                        
+                }
 
                 // Check if a file switch is required
                 if (pAudioData->switchFiles)
@@ -1512,7 +1531,7 @@ void vorbis_read_pcm_frames(ma_data_source *pDataSource, void *pFramesOut, ma_ui
                         break; // Exit the loop after the file switch
                 }
 
-                ma_libvorbis *decoder = getCurrentVorbisDecoder();                
+                ma_libvorbis *decoder = getCurrentVorbisDecoder();
 
                 if ((getCurrentImplementationType() != VORBIS && !isSkipToNext()) || (decoder == NULL))
                 {
