@@ -6,6 +6,7 @@
 #include <miniaudio.h>
 #include <miniaudio_libopus.h>
 #include <miniaudio_libvorbis.h>
+#include <sys/wait.h>
 #include "m4a.h"
 #include <stdatomic.h>
 #include <stdbool.h>
@@ -132,6 +133,7 @@ typedef struct
         char addToMainPlaylist[6];
         char updateLibrary[6];
         char quit[6];
+        char hardQuit[6];
         char hardSwitchNumberedSong[6];
         char hardPlayPause[6];
         char hardPrev[6];
@@ -143,8 +145,10 @@ typedef struct
         char hardShowKeys[6];
         char hardShowKeysAlt[6];
         char hardEndOfPlaylist[6];
-        char hardShowLibrary[6];        
+        char hardShowLibrary[6];
         char hardShowLibraryAlt[6];
+        char hardShowSearch[6];
+        char hardShowSearchAlt[6];        
         char hardShowTrack[6];
         char hardShowTrackAlt[6];
         char hardNextPage[6];
@@ -177,7 +181,8 @@ typedef enum {
     SONG_VIEW,
     KEYBINDINGS_VIEW,
     PLAYLIST_VIEW,
-    LIBRARY_VIEW
+    LIBRARY_VIEW,
+    SEARCH_VIEW
 } ViewState;
 
 typedef struct {
@@ -330,6 +335,8 @@ void activateSwitch(AudioData *pPCMDataSource);
 
 void executeSwitch(AudioData *pPCMDataSource);
 
+gint64 getLengthInSec(double duration);
+
 int displaySongNotification(const char *artist, const char *title, const char *cover);
 
 int getCurrentVolume(void);
